/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of gnome-spell bonobo component
    copied from echo.c written by Miguel de Icaza and updated for Spell.idl needs

    Copyright (C) 1999, 2000 Helix Code, Inc.
    Authors:                 Radek Doulik <rodo@helixcode.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include <config.h>
#include <glib.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-config.h>
#include <bonobo.h>

#include "Spell.h"
#include "dictionary.h"

#include <enchant.h>

typedef struct {
	EnchantBroker * config;
	EnchantDict   * speller;
} SpellEngine;

static BonoboObjectClass                  *dictionary_parent_class;

#define DICT_DEBUG(x)

static void release_engines (GNOMESpellDictionary *dict);
extern const GSList   *spell_checker_get_available_languages (void);

static gchar *
engine_to_language (GNOMESpellDictionary *dict, SpellEngine *se)
{
	return (gchar *) se ? g_hash_table_lookup (dict->engines_ht, se) : NULL;
}

static SpellEngine *
language_to_engine (GNOMESpellDictionary *dict, gchar *language)
{
	return (SpellEngine *) language ? g_hash_table_lookup (dict->engines_ht, language) : NULL;
}

static void
raise_error (CORBA_Environment * ev, const gchar *s)
{
	GNOME_Spell_Dictionary_Error *exception;
	exception = GNOME_Spell_Dictionary_Error__alloc ();
		
	exception->error = CORBA_string_dup (s);
	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
			     ex_GNOME_Spell_Dictionary_Error,
			     exception);
}

static void
gnome_spell_dictionary_init (GObject *object)
{
	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (object);

	dict->engines = NULL;
	dict->languages = g_hash_table_new (g_str_hash, g_str_equal);
	dict->engines_ht = g_hash_table_new (NULL, NULL);
}

static char **
dup_string_list (char ** str_list, size_t list_len)
{
	char ** new_str_list;
	size_t i;

	new_str_list = g_new0 (char *, list_len + 1);
	for (i = 0; i < list_len; i++)
		new_str_list [i] = g_strdup (str_list [i]);

	return new_str_list;
}

static void
dictionary_finalize (GObject *object)
{
	GNOMESpellDictionary *dictionary = GNOME_SPELL_DICTIONARY (object);

	release_engines (dictionary);
	g_hash_table_destroy (dictionary->languages);
	dictionary->languages = NULL;
	g_hash_table_destroy (dictionary->engines_ht);
	dictionary->engines_ht = NULL;

	G_OBJECT_CLASS (dictionary_parent_class)->finalize (object);
}

static SpellEngine *
new_engine (const gchar *language, CORBA_Environment *ev)
{
	SpellEngine *se;

	se = g_new0 (SpellEngine, 1);
	se->config = enchant_broker_init ();
	se->speller = enchant_broker_request_dict (se->config, language);

	if(se->speller == NULL)
		raise_error (ev, enchant_broker_get_error (se->config));

	return se;
}

static gboolean
remove_language (gpointer key, gpointer value, gpointer user_data)
{
	g_free (key);

	return TRUE;
}

static gboolean
remove_engine_ht (gpointer key, gpointer value, gpointer user_data)
{
	g_free (value);

	return TRUE;
}

static void
release_engines (GNOMESpellDictionary *dict)
{
	for (; dict->engines; ) {
		SpellEngine *se = dict->engines->data;

		if (se->speller)
			enchant_broker_free_dict (se->config, se->speller);
		if (se->config)
			enchant_broker_free (se->config);
		g_free (se);
		dict->engines = g_slist_remove (dict->engines, se);
	}

	g_hash_table_foreach_remove (dict->languages, remove_language, NULL);
	g_hash_table_foreach_remove (dict->languages, remove_engine_ht, NULL);

	dict->engines = NULL;
}

static GNOME_Spell_LanguageSeq *
impl_gnome_spell_dictionary_get_languages (PortableServer_Servant servant, CORBA_Environment *ev)
{
	GNOME_Spell_LanguageSeq *seq;
	GSList *l, *langs;
	gint i, pos;

        langs = spell_checker_get_available_languages();

	seq = GNOME_Spell_LanguageSeq__alloc ();
	seq->_length = g_slist_length(langs);

	if (seq->_length == 0)
		return seq;

	seq->_buffer = CORBA_sequence_GNOME_Spell_Language_allocbuf (seq->_length);

        i = 0;
        while (langs != NULL )
        {
                const LangInfo*l = (const LangInfo*)langs->data;
                seq->_buffer [i].name = CORBA_string_dup (l->name);
                seq->_buffer [i].abbreviation = CORBA_string_dup (l->abbreviation);
                i++;
                langs = g_slist_next(langs);
        }

	CORBA_sequence_set_release (seq, CORBA_TRUE);
	g_slist_free (langs);

	return seq;
}

static void
impl_gnome_spell_dictionary_set_language (PortableServer_Servant  servant,
					  const CORBA_char       *language,
					  CORBA_Environment      *ev)
{
	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
	const gchar *s, *begin, *end;
	gchar *one_language;
	gint len;

#ifndef G_DISABLE_CHECKS
	g_return_if_fail (dict);
#else
	if (!dict)
		return;
#endif
	if (!language)
		language = "";

	DICT_DEBUG (printf ("setLanguage: %s\n", language));

	release_engines (dict);
	for (s = language; *s; s = end) {
		begin = s;
		while (*begin && *begin == ' ')
			begin++;
		end = begin;
		len = 0;
		while (*end && *end != ' ') {
			end ++;
			len ++;
		}

		if (len) {
			SpellEngine *se;
			
			one_language = g_strndup (begin, len);
			se = new_engine (one_language, ev);
			dict->engines = g_slist_prepend (dict->engines, se);
			g_hash_table_insert (dict->languages, one_language, se);
			g_hash_table_insert (dict->engines_ht, se, g_strdup (one_language));
		}
	}
}

static CORBA_boolean
engine_check_word (SpellEngine *se, const gchar *word, CORBA_Environment *ev)
{
	CORBA_boolean result = CORBA_FALSE;
	gint enchant_result;

#ifndef G_DISABLE_CHECKS
	g_return_val_if_fail (se->speller, CORBA_TRUE);
#else
	if (!se->speller)
		return CORBA_TRUE;
#endif
	enchant_result = enchant_dict_check (se->speller, word, strlen (word));
	if (enchant_result == 0)
		result = CORBA_TRUE;
	if (enchant_result == -1) {
		g_warning ("enchant error: %s\n", enchant_dict_get_error (se->speller));
		raise_error (ev, enchant_dict_get_error (se->speller));
	}

	return result;
}

static CORBA_boolean
impl_gnome_spell_dictionary_check_word (PortableServer_Servant servant, const CORBA_char *word, CORBA_Environment *ev)
{
	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
	CORBA_boolean result = CORBA_FALSE;
	GSList *l;
	gboolean valid_speller = FALSE;

#ifndef G_DISABLE_CHECKS
	g_return_val_if_fail (word, result);
#else
	if (!word)
		return result;
#endif
	if (!strcmp (word, "Ximian"))
		return CORBA_TRUE;

	for (l = dict->engines; l; l = l->next) {
		if (((SpellEngine *) l->data)->speller) {
			valid_speller = TRUE;
			if (engine_check_word ((SpellEngine *) l->data, word, ev))
				result = CORBA_TRUE;
		}
	}

	if (!valid_speller) {
		DICT_DEBUG (printf ("Dictionary check_word: %s --> 1 (not valid speller)\n", word));
		return CORBA_TRUE;
	}
	DICT_DEBUG (printf ("Dictionary check_word: %s --> %d\n", word, result));
	return result;
}

static void
impl_gnome_spell_dictionary_add_word_to_session (PortableServer_Servant servant, const CORBA_char *word, CORBA_Environment *ev)
{
	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
	GSList *l;

#ifndef G_DISABLE_CHECKS
	g_return_if_fail (word);
#else
	if (!word)
		return;
#endif
	DICT_DEBUG (printf ("Dictionary add_word_to_session: %s\n", word));
	for (l = dict->engines; l; l = l->next) {
		if (((SpellEngine *) l->data)->speller)
			enchant_dict_add_to_session (((SpellEngine *) l->data)->speller, word, strlen (word));
	}
}

static void
impl_gnome_spell_dictionary_add_word_to_personal (PortableServer_Servant servant,
						  const CORBA_char *word, const CORBA_char *language, CORBA_Environment *ev)
{
	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
	SpellEngine *se;

#ifndef G_DISABLE_CHECKS
	g_return_if_fail (word && language);
#else
	if (!word || !language)
		return;
#endif
	DICT_DEBUG (printf ("Dictionary add_word_to_personal: %s (%s)\n", word, language));
	se = (SpellEngine *) g_hash_table_lookup (dict->languages, language);

	if (se && se->speller) {
		enchant_dict_add_to_pwl (se->speller, word, strlen (word));
		DICT_DEBUG (printf ("Added and saved.\n"));
	}
}

static void
impl_gnome_spell_dictionary_set_correction (PortableServer_Servant servant,
					    const CORBA_char *word, const CORBA_char *replacement, const CORBA_char *language, CORBA_Environment *ev)
{
	GNOMESpellDictionary *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
	SpellEngine *se;

#ifndef G_DISABLE_CHECKS
	g_return_if_fail (word && replacement);
#else
	if (!word || !replacement)
		return;
#endif
	DICT_DEBUG (printf ("Dictionary correction: %s <-- %s\n", word, replacement));
	se = (SpellEngine *) g_hash_table_lookup (dict->languages, language);

	if (se && se->speller) {
		enchant_dict_store_replacement (se->speller, word, strlen (word),
						replacement, strlen (replacement));
		DICT_DEBUG (printf ("Set and saved.\n"));
	}
}

static GNOME_Spell_StringSeq *
impl_gnome_spell_dictionary_get_suggestions (PortableServer_Servant servant,
					     const CORBA_char *word, CORBA_Environment *ev)
{
	GNOMESpellDictionary  *dict = GNOME_SPELL_DICTIONARY (bonobo_object_from_servant (servant));
	char **suggestions;
	size_t number_of_suggestions;
	GNOME_Spell_StringSeq *seq = NULL;
	GSList *l, *suggestion_list = NULL;
	gint i, len, pos;

#ifndef G_DISABLE_CHECKS
	g_return_val_if_fail (word, NULL);
#else
	if (!word)
		return NULL;
#endif
	DICT_DEBUG (printf ("Dictionary correction: %s\n", word));

	len = 0;
	for (l = dict->engines; l; l = l->next) {
		SpellEngine *se = (SpellEngine *) l->data;

		if (se->speller) {
			suggestions = enchant_dict_suggest (se->speller, word, strlen (word), &number_of_suggestions);
			suggestion_list = g_slist_prepend (suggestion_list, 
							   (gpointer) dup_string_list (suggestions, number_of_suggestions));
			len += 2*number_of_suggestions;
			suggestion_list = g_slist_prepend (suggestion_list, engine_to_language (dict, se));
			suggestion_list = g_slist_prepend (suggestion_list, GINT_TO_POINTER (number_of_suggestions));
			enchant_dict_free_string_list (se->speller, suggestions);
		}
	}

	seq          = GNOME_Spell_StringSeq__alloc ();
	seq->_length = len;

	if (seq->_length == 0)
		return seq;

	seq->_buffer = CORBA_sequence_CORBA_string_allocbuf (seq->_length);

	pos = 0;
	for (l = suggestion_list; l; l = l->next) {
		gint list_len;
		gchar *language;

		list_len = GPOINTER_TO_INT (l->data);
		l = l->next;
		language = (gchar *) l->data;
		l = l->next;
		suggestions = (char **) l->data;
		for (i = 0; i < list_len; i ++, pos ++) {
			seq->_buffer [pos] = CORBA_string_dup (suggestions [i]);
			pos ++;
			seq->_buffer [pos] = CORBA_string_dup (language);
		}

		g_strfreev (suggestions);
	}
	CORBA_sequence_set_release (seq, CORBA_TRUE);
	g_slist_free (suggestion_list);

	return seq;
}

/*
 * If you want users to derive classes from your implementation
 * you need to support this method.
 */
POA_GNOME_Spell_Dictionary__epv *
gnome_spell_dictionary_get_epv (void)
{
	POA_GNOME_Spell_Dictionary__epv *epv;

	epv = g_new0 (POA_GNOME_Spell_Dictionary__epv, 1);


	return epv;
}

static void
gnome_spell_dictionary_class_init (GNOMESpellDictionaryClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	POA_GNOME_Spell_Dictionary__epv *epv = &klass->epv;

	dictionary_parent_class   = g_type_class_ref (BONOBO_TYPE_OBJECT);
	object_class->finalize    = dictionary_finalize;

	epv->getLanguages         = impl_gnome_spell_dictionary_get_languages;
	epv->setLanguage          = impl_gnome_spell_dictionary_set_language;
	epv->checkWord            = impl_gnome_spell_dictionary_check_word;
	epv->addWordToSession     = impl_gnome_spell_dictionary_add_word_to_session;
	epv->addWordToPersonal    = impl_gnome_spell_dictionary_add_word_to_personal;
	epv->getSuggestions       = impl_gnome_spell_dictionary_get_suggestions;
	epv->setCorrection        = impl_gnome_spell_dictionary_set_correction;
}

BONOBO_TYPE_FUNC_FULL (
	GNOMESpellDictionary,          /* Glib class name */
	GNOME_Spell_Dictionary,        /* CORBA interface name */
	BONOBO_TYPE_OBJECT,            /* parent type */
	gnome_spell_dictionary);       /* local prefix ie. 'echo'_class_init */

BonoboObject *
gnome_spell_dictionary_new (void)
{
	return g_object_new (GNOME_SPELL_DICTIONARY_TYPE, NULL);
}
